/*   video test program for the AM-CLR module  
 *   free public release
 *
 *   Werner Jaeger        email: e9026603@student.tuwien.ac.at
 *   Wilfried Kubinger    email: wk@flexaut.tuwien.ac.at
 *
 *   Institute of Flexible Automation
 *   University of Technology Vienna
 *   Austria, Europe
 *
 *   test_icpci_clr.c  v1.0  1998/01/05
 */

/* Change History:

   1999/02/19 14:01:46  matthias
   changes to match the new on-board/upper memory mapping scheme (untested)
   struct icp_info is only available in bma-mode

   1999/03/02 10:21:31  matthias
   minor adjustments to avoid egcs111 warnings

   1999/05/03 12:47:07 matthias
   ajustments to version 0.5.0 (icpciMain.h > icpciExt.h)
   card 8/16/32 -> u_char/short/long

   1999/09/21 15:00:16 matthias
   ajustments for version 0.5.1
   the statistics are now u_long values
   due to the capability of decimazation we have to distinguish between
   MEMWIDTH/HEIGHT and BMAWIDTH/HEIGHT
   RAWSIZE is now MEMBUFSIZE ans PAGEDSIZE is now BMABUFSIZE
   MEMSIZE is renamed to MEMMAPSIZE

 */


/*
 * Defines:
 */

#define FONT_NAME "9x15"

/*
 * Includes:
 */

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include <asm/page.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#include "icpciExt.h"

/*
 * Typedefs:
 */

typedef enum {
  PSEUDO_8BIT,
  RGB_16BIT,
  BGR_16BIT,
  RGB_32BIT,
  BGR_32BIT
} convert_mode_t;

typedef enum {
  READ_MODE,
  MMAP_MODE,
  BMA_MODE
} operation_mode_t;

typedef enum {
  SIGNAL_MODE,
  SELECT_MODE
} synchronisation_mode_t;

typedef enum {
  GRAB_MODE,
  SNAP_MODE,
  FREEZE_MODE
} acquisition_mode_t;

typedef enum {
  NORMAL_MODE,
  FIELDS_MODE,
  AOI_MODE
} grab_mode_t;

/*
 * Devices:
 */

const char *c1_camera_device[] = {
  "/dev/ic0c10" , "/dev/ic0c11" , "/dev/ic0c12" , "/dev/ic0c13" ,
  "/dev/ic1c10" , "/dev/ic1c11" , "/dev/ic1c12" , "/dev/ic1c13" ,
  "/dev/ic2c10" , "/dev/ic2c11" , "/dev/ic2c12" , "/dev/ic2c13" ,
  "/dev/ic3c10" , "/dev/ic3c11" , "/dev/ic3c12" , "/dev/ic3c13"
};

const char *c1_ioctl_device[] = {
  "/dev/ic0c1I",
  "/dev/ic1c1I",
  "/dev/ic2c1I",
  "/dev/ic3c1I"
};

const char *c1_ident_device[] = {
  "/dev/ic0idI",
  "/dev/ic1idI",
  "/dev/ic2idI",
  "/dev/ic3idI" 
};

/*
 * Globals:
 */

extern char *optarg;

char colortable[64];

convert_mode_t          convert_mode;
operation_mode_t        operation_mode = READ_MODE;
synchronisation_mode_t  synchronisation_mode = SIGNAL_MODE;
acquisition_mode_t      acquisition_mode = GRAB_MODE;
grab_mode_t             grab_mode = NORMAL_MODE;

/*
 * Prototypes:
 */

int log_amclr (int, int);
int fpga_load (struct c1_fpga_t *);
int ascii2decimal (char *);
void rgb2image (XImage *, u_long *, int, int);
void rgb_ilace2image (XImage *, u_long *, int, int);

/*
 * Main:
 */

int main(int argc, char **argv) {

  Display *display;
  Visual *visual;
  Window parent, window;
  GC context;
  XGCValues values;
  XImage *image;
  XSetWindowAttributes attributes;
  XEvent event;
  XColor color;
  Colormap colormap;
  KeySym keysym;
  int screen, depth, bitmap_pad;
  unsigned int mask;

  fd_set wait_fdset;
  struct timeval wait_time;
  struct timeval tv_start, tv_sample;
  struct timezone tz;
  unsigned long frame_count;
  float count_time;
  char text[32] = "";

  struct c1_fpga_t fpga;
  struct c1_geometry_t geometry;
  /* union icp_dev_t cam_cnf; */
  struct icp_info_t *icp_info = NULL;

  int option, board_arg, camera_arg;

  int width=0, height=0, rawsize=0, pagedsize=0;
  int i, red, green, blue, show;
  int framegrabber=0;
  u_long *read_buffer=NULL, *image_mmaped=NULL;
  size_t mem_size;

  /* --- command line parsing --------------------------------------------- */

  /* default options */
  board_arg = 0;
  camera_arg = 2;
  strcpy (text, "empty");  /* empty FPGA-Chip */

  /* get options */
  while ((option = getopt (argc, argv, "o:b:c:l:")) != -1) {
    switch (option) {
    case 'o':
      if (!strcmp ("read", optarg)) operation_mode = READ_MODE;
      else if (!strcmp ("mmap", optarg)) operation_mode = MMAP_MODE;
      else if (!strcmp ("bma", optarg)) operation_mode = BMA_MODE;
      else {
        fprintf (stderr, "wrong operation mode \"%s\".\n", optarg);
        exit (-1);
      }
      break;
    case 'b':
      if (!strcmp ("0", optarg)) board_arg = 0;
      else if (!strcmp ("1", optarg)) board_arg = 1;
      else if (!strcmp ("2", optarg)) board_arg = 2;
      else if (!strcmp ("3", optarg)) board_arg = 3;
      else {
        fprintf (stderr, "wrong board number \"%s\".\n", optarg);
        exit (-1);
      }
      break;
    case 'c':
      if (!strcmp ("0", optarg)) camera_arg = 0;
      else if (!strcmp ("1", optarg)) camera_arg = 1;
      else if (!strcmp ("2", optarg)) camera_arg = 2;
      else if (!strcmp ("3", optarg)) camera_arg = 3;
      else {
        fprintf (stderr, "wrong camera number \"%s\".\n", optarg);
        exit (-1);
      }
      break;
    case 'l':
      strcpy (text, optarg);
      break;
    default:
      fprintf (stderr, "Usage: -o operation mode (read) [read|mmap|bma]\n");
      fprintf (stderr, "       -b board number (0) [0,1,2,3]\n");
      fprintf (stderr, "       -c camera number (2) [0,1,2,3]\n");
      fprintf (stderr, "       -l load FPGA-Chip [mcs file name]\n");
      exit (-1);
    }
  }

  /* --- setup framegrabber device ---------------------------------------- */

  switch (operation_mode) {
  case READ_MODE:
    if ((framegrabber = open (c1_camera_device[board_arg * 4 + camera_arg],
                              O_RDONLY)) < 0) {
      perror ("open framegrabber device (with AM-CLR)");
      exit (-1);
    }
    break;

  case MMAP_MODE:
    if ((framegrabber = open (c1_ioctl_device[board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      exit (-1);
    }
    break;

  case BMA_MODE:
    if ((framegrabber = open (c1_ident_device[board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      exit (-1);
    }
  }

  /* get the current loaded FPGA-Timing */
  if (ioctl (framegrabber, GIOC_C1_GET_FPGA, &geometry) < 0) {
    perror ("GIOC_C1_GET_FPGA");
    exit (-1);
  }
  /* check if an FPGA-Timing was given at command line */
  if (strcmp ("empty", text) != 0) {
    /* check if the new FPGA-Timing is different from the current loaded */
    if (strcmp (text, geometry.mcs_name) != 0) {
      strcpy (fpga.fpga_name, text);
      /* read the new MCS-Timing File */
      if (fpga_load (&fpga) < 0) {
        fprintf (stderr, "can't load mcs file.\n");
        exit (-1);
      }
      /* write the new Timing into FPGA-Chip */
      if (ioctl (framegrabber, GIOC_C1_SET_FPGA, &fpga) < 0) {
        perror ("GIOC_C1_SET_FPGA");
        exit (-1);
      }
    }
  } else {
    /* no command line argument, check if a valid FPGA-Timing is loaded */
    if (strcmp ("empty", geometry.mcs_name) == 0) {
      fprintf (stderr, "no timing loaded into FPGA-Chip.\n");
      exit (-1);
    }
  }

  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (ioctl (framegrabber, GIOC_GE_SET_CAMERA, &camera_arg) < 0) {
      perror ("GIOC_GE_SET_CAMERA");
      exit (-1);
    }
  }

#if(0)
  /* read, change and write-back Camera-Configuration */

  if (ioctl (framegrabber, GIOC_GE_GET_CAMCNF, &cam_cnf) < 0) {
    perror ("GIOC_GE_GET_CAMCNF");
    exit (-1);
  }

  /* her you can change the default AM-CLR settings */

  if (ioctl (framegrabber, GIOC_GE_SET_CAMCNF, &cam_cnf) < 0) {
    perror ("GIOC_GE_SET_CAMCNF");
    exit (-1);
  }
#endif

  /* this test program works only if the AM-CLR is set to  TRUECOLOR mode */
  if (ioctl (framegrabber, GIOC_C1_GET_COLORMODE, &i) < 0) {
    perror ("GIOC_C1_GET_COLORMODE");
    exit (-1);
  }
  if (i != 0) {
    fprintf (stderr, "AM-CLR color mode in not TRUECOLOR.\n");
    exit (-1);
  }

  /* --- get image dimensions --------------------------------------------- */

  switch (operation_mode) {
  case READ_MODE:
  case MMAP_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_MEMWIDTH, &width) < 0) {
      perror ("GIOC_GE_GET_WIDTH");
      exit (-1);
    }
    if (ioctl (framegrabber, GIOC_GE_GET_MEMHEIGHT, &height) < 0) {
      perror ("GIOC_GE_GET_HEIGHT");
      exit (-1);
    }
    break;
    
  case BMA_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_BMAWIDTH, &width) < 0) {
      perror ("GIOC_GE_GET_WIDTH");
      exit (-1);
    }
    if (ioctl (framegrabber, GIOC_GE_GET_BMAHEIGHT, &height) < 0) {
    perror ("GIOC_GE_GET_HEIGHT");
    exit (-1);
    }
    break;
  }
  
  if (ioctl (framegrabber, GIOC_GE_GET_MEMBUFSIZE, &rawsize) < 0) {
    perror ("GIOC_GE_GET_RAWSIZE");
    exit (-1);
  }
  if (ioctl (framegrabber, GIOC_GE_GET_BMABUFSIZE, &pagedsize) < 0) {
    perror ("GIOC_GE_GET_PAGEDSIZE");
    exit (-1);
  }

  /* --- initialize the given operation mode ------------------------------ */

  switch (operation_mode) {
  case READ_MODE:
    if ((read_buffer = (u_long *) malloc (rawsize)) == NULL) {
      fprintf (stderr, "can't malloc read buffer.\n");
      exit (-1);
    }
    break;
  case MMAP_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_MEMMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_MEMSIZE");
      exit (-1);
    }
    if ((image_mmaped = (u_long*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				       framegrabber , 0)) == (void*)-1) {
      perror ("Can't map camera device");
      exit (-1);
    }
    break;
    
  case BMA_MODE:
    if (ioctl (framegrabber, GIOC_GE_GET_BMAMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_BMASIZE");
      exit (-1);
    }
    if ((image_mmaped = (void*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				     framegrabber, 0)) == (void*)-1) {
      perror ("Can't map camera device");
      exit (-1);
    }
    /* pointer to the info structure of the acquired image */
    icp_info = (struct icp_info_t *)(((u_char *)image_mmaped) + rawsize);
    break;
  }

  /* --- setup X11 -------------------------------------------------------- */

  if ((display = XOpenDisplay (NULL)) == NULL) {
    fprintf (stderr, "can't open X11 display.\n");
    exit (-1);
  }

  screen = DefaultScreen (display);
  depth = DefaultDepth (display, screen);
  visual = DefaultVisual (display, screen);
  colormap = DefaultColormap (display, screen);
  parent = RootWindow (display, screen);

  switch (depth) {
  case 8:
    /* 8 Bit PseudoColor Display, alloc color cells */
    i=0;
    for (blue=0; blue<4; blue++) {
      color.blue = 65535*blue/3;
      for (green=0; green<4; green++) {
        color.green = 65535*green/3;
        for (red=0; red<4; red++) {
          color.red = 65535*red/3;
          color.pad = 0;
          if (XAllocColor (display, colormap, &color) == 0) {
            printf ("can't allocate 64 colors (only %d).\n", i);
            exit (-1);
          }
          colortable[i++] = color.pixel;
        }
      }
    }
    bitmap_pad = 8;
    break;
  case 16:
    /* 16 Bit TrueColor Display, use color mask */
    bitmap_pad = 16;
    break;
  case 24:
    /* 24 Bit TrueColor Display, use color mask */
    bitmap_pad = 32;
    break;
  default:
    fprintf (stderr, "need a X11 display with depth 8, 16 or 24.\n");
    exit (-1);
    break;
  }

  window = XCreateWindow (display, parent, 0, 0, width, height, 4, depth,
                          InputOutput, visual, 0, &attributes);  

  context = XCreateGC (display, window, 0, &values);

  image = XCreateImage (display, visual, depth, ZPixmap, 0, NULL, 
                        width, height, bitmap_pad, width*bitmap_pad/8);

  if ((image->data = (char *) malloc ((rawsize/4)*(bitmap_pad/8))) == NULL) {
    fprintf (stderr, "can't malloc image data.\n");
    exit (-1);
  }

  /* select the proper convert mode for the available X11 display  */
  switch (depth) {
    case 8:
      convert_mode = PSEUDO_8BIT;
      break;
    case 16:
      if (image->red_mask > image->blue_mask) {
        convert_mode = RGB_16BIT;
      } else {
        convert_mode = BGR_16BIT;
      }
      break;
    case 24:
      if (image->red_mask > image->blue_mask) {
        convert_mode = RGB_32BIT;
      } else {
        convert_mode = BGR_32BIT;
      }
      break;
  }

  /* --- show frames in window -------------------------------------------- */

  mask = ExposureMask | KeyPressMask | ButtonPressMask;
  XSelectInput (display, window, mask);
  XMapWindow (display, window); 
  XStoreName (display, window, "*** Grap Image ***"); 

  show = 1;
  frame_count = 0;
  gettimeofday (&tv_start, &tz);  

  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (ioctl (framegrabber, GIOC_CR_SET_GRAB, NULL) < 0) {
      perror ("GIOC_CR_SET_GRAB");
      exit (-1);
    }
  }

  while (show) {

    switch (acquisition_mode) {

    case SNAP_MODE:
    case FREEZE_MODE: 
      XWindowEvent (display, window, mask, &event);
      break;

    case GRAB_MODE:
      switch(operation_mode) {
   
      case READ_MODE:
        if (read (framegrabber, read_buffer, rawsize) < 0) {
          if (errno == ENODATA)
              XStoreName (display, window,  "*** No Sync ***");
          else {
            perror ("read framegrabber device");
            exit (-1);
          }        
        }
        rgb2image (image, read_buffer, width, height);
        break;

      case MMAP_MODE:
        FD_ZERO (&wait_fdset);
        FD_SET (framegrabber, &wait_fdset);	    
        wait_time.tv_sec = 1;
        wait_time.tv_usec = 0;
        switch (select (getdtablesize(), &wait_fdset, NULL, NULL, 
                        &wait_time)) {
        case -1:
	  XStoreName (display, window, "*** Signal in select ***\n");
          break;
        case 0:
          XStoreName (display, window, "*** No Sync ***\n");
          break;
        case 1:
	  break;
        default:
          perror ("select got unexpected return value");
          exit (-1);
        }
        rgb_ilace2image (image, image_mmaped, width, height);
        break;

      case BMA_MODE:
        FD_ZERO (&wait_fdset);
        FD_SET (framegrabber, &wait_fdset);	    
        wait_time.tv_sec = 1;
        wait_time.tv_usec = 0;
        switch (select (getdtablesize(), &wait_fdset, NULL, NULL, 
                        &wait_time)) {
        case -1:
	  XStoreName (display, window, "*** Signal in select ***\n");
          break;
        case 0:
          XStoreName (display, window, "*** No Sync ***\n");
          break;
        case 1:
	  break;
        default:
	  perror ("select got unexpected return value");
	  exit (-1);
        }
        rgb2image (image, image_mmaped, width, height);
        break;
  
      default:
        fprintf (stderr, "unknown acquisition mode.\n");
        exit (-1);
      }
       
      XPutImage (display, window, context, image, 0, 0, 0, 0, width, height);

      frame_count++;
      gettimeofday (&tv_sample, &tz);
      count_time = tv_sample.tv_sec - tv_start.tv_sec;
      if (tv_start.tv_usec<tv_sample.tv_usec) {
        count_time += 1e-6 * (tv_sample.tv_usec - tv_start.tv_usec);
      } else {
        count_time += 1e-6 * (tv_start.tv_usec - tv_sample.tv_usec); 
      } 
      sprintf (text, "%4.1f  Frames/sec", frame_count/count_time);
      XStoreName (display, window, text);
      if (False == XCheckWindowEvent (display, window, mask, &event)) continue;
      break;
 
    default:
      fprintf (stderr, "unknown operation mode.\n");
      exit (-1);
    }

    switch (event.type) {
    case Expose:
      if (event.xexpose.count != 0) break;
      XPutImage (display, window, context, image, 0, 0, 0, 0, width, height);
      break;
    case ButtonPress:
      show = 0;
      break; 
    case KeyPress:
      keysym = XKeycodeToKeysym (display, event.xkey.keycode, 0);
      switch (keysym) {  
      case XK_q: 
        show = 0; 
        break;
      case XK_s: 
        break;
      case XK_g:;
        frame_count = 0;
        gettimeofday (&tv_start, &tz);
        if (operation_mode != READ_MODE) {
          /* MMAP_MODE or BMA_MODE */ 
          if (ioctl (framegrabber, GIOC_CR_SET_GRAB, NULL) < 0) {
            perror ("GIOC_CR_SET_GRAB");
            exit (-1);
          }
        }
        XStoreName (display, window, "*** Grap Image ***");
        acquisition_mode = GRAB_MODE;
        break;
      case XK_f:
        if (operation_mode != READ_MODE) {
          /* MMAP_MODE or BMA_MODE */ 
          if (ioctl (framegrabber, GIOC_CR_SET_FREEZE, NULL) < 0) {
            perror ("GIOC_CR_SET_FREZZE");
            exit (-1);
          }
        }
        XStoreName (display, window, "*** Freeze Image ***");
        acquisition_mode = FREEZE_MODE;
        break;
      case XK_F1:
        /* log everything */
        if (log_amclr (framegrabber, 1) < 0) exit (-1);
        break;
      case XK_F2:
        /* Register */
        if (log_amclr (framegrabber, 2) < 0) exit (-1);
        break;
      case XK_F3:
        /* Geometry */
        if (log_amclr (framegrabber, 3) < 0) exit (-1);
        break;
      case XK_F4:
        /* Colormode */
        if (log_amclr (framegrabber, 4) < 0) exit (-1);
        break;
      case XK_F5:
        /* Camera Configuration */
        if (log_amclr (framegrabber, 5) < 0) exit (-1);
        break;
      case XK_F6:
        /* Camera Port */
        if (log_amclr (framegrabber, 6) < 0) exit (-1);
        break;
      case XK_F7:
        /* Reference Voltage */
        if (log_amclr (framegrabber, 7) < 0) exit (-1);
        break;
      case XK_F8:
        /* Color Space Converter */
        if (log_amclr (framegrabber, 8) < 0) exit (-1);
        break;
      case XK_F9:
        /* Input Lookup Table */
        if (log_amclr (framegrabber, 9) < 0) exit (-1);
        break;
      case XK_F10:
        /* Output Lookup Table */
        if (log_amclr (framegrabber, 10) < 0) exit (-1);
        break;
      }
    }
  }

  /* --- close X11 -------------------------------------------------------- */

  XDestroyImage (image);
  XFreeGC (display, context);
  XCloseDisplay (display);

  /* --- close framegrabber device ---------------------------------------- */

  if (operation_mode != READ_MODE) {
    /* MMAP_MODE or BMA_MODE */
    if (ioctl (framegrabber, GIOC_CR_SET_FREEZE, NULL) < 0) {
      perror ("GIOC_CR_SET_FREEZE");
      exit (-1);
    }
    if (operation_mode == BMA_MODE) {
      /* output info of last acquired image */
      printf ("clock tics: %ld\n", icp_info->clock_tics);
      printf ("time stamp: %ld sec, %ld usec\n", 
	      icp_info->time_stamp.tv_sec, icp_info->time_stamp.tv_usec);
      printf ("frames captured: %ld\n", icp_info->frame_nums.frames_captured);
      printf ("   frames valid: %ld\n", icp_info->frame_nums.frames_valid);
      printf ("    frames lost: %ld\n", icp_info->frame_nums.frames_lost);
    }
  }

  if (read_buffer != NULL) free (read_buffer);

  if (operation_mode != READ_MODE) {
    if (munmap ((char*)image_mmaped, mem_size) < 0) {
      perror ("unmap framebuffer");
      exit (-1);
    }
  } 

  if (close (framegrabber) < 0) {
    perror ("close framegrabber device");
    exit (-1);
  }
  return (0);
}
 

/*
 * Functions used only in test_icpci_clr.c
 */

/*
 * int log_amclr (int, int)
 *
 * write the return values of the different 'GIOC_C1_GET_'-Ioctls,
 * the 'GIOC_GET_CAMERA' and the 'GIOC_GET_CAMCNF'-Ioctl to the
 * log-file 'am-clr.log'
 */

/*    ioctl                  |   log_action
 *  -------------------------+---------------
 *    all                    |       1
 *    GIOC_C1_GET_REGISTER   |       2
 *    GIOC_C1_GET_FPGA       |       3
 *    GIOC_C1_GET_COLORMODE  |       4
 *    GIOC_GE_GET_CAMCNF     |       5
 *    GIOC_GE_GET_CAMERA     |       6
 *    GIOC_C1_GET_REF        |       7
 *    GIOC_C1_GET_CSC        |       8
 *    GIOC_C1_GET_ILUT       |       9
 *    GIOC_C1_GET_OLUT       |       10
 */

int 
log_amclr (int framegrabber, int log_action) {

  FILE *log_file;
  int error_count, i, k;

  struct timeval tv;
  struct timezone tz;

  struct c1_geometry_t geometry;
  struct c1_regs_t regs;
  struct c1_ref_t ref;
  struct c1_csc_t csc;
  struct c1_ilut_t ilut;
  struct c1_olut_t olut;
  union icp_dev_t camcnf;
  int value;

  if ((log_file = fopen ("amclr.log", "a")) == NULL) {
    perror ("cannot open amclr.log");
    return (-1);
  }  
  error_count = 0;

  gettimeofday (&tv, &tz);
  fprintf (log_file, "\n###\n### %ld sec : %ld usec\n###\n\n", 
    tv.tv_sec, tv.tv_usec); 

  switch (log_action) {

    /* write everything to the log-file */
  case 1:

    /* write current AM-CLR register settings to the log-file */
  case 2:
    if (ioctl (framegrabber, GIOC_C1_GET_REGISTER, &regs) < 0) {
      perror ("GIOC_C1_GET_REGISTER");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Register)\n");
      error_count++;
    } else {
      fprintf (log_file," MODID: 0x%02x   %03d\n", 
               regs.modid_reg, regs.modid_reg);
      fprintf (log_file,"CAMCON: 0x%02x   %03d\n", 
               regs.camcon_reg, regs.camcon_reg);
      fprintf (log_file,"MODCON: 0x%02x   %03d\n", 
               regs.modcon_reg, regs.modcon_reg);
      fprintf (log_file,"ADCCON: 0x%02x   %03d\n",
               regs.adccon_reg, regs.adccon_reg);
      fprintf (log_file," RNREF: 0x%02x   %03d\n", 
               regs.rnref_reg, regs.rnref_reg);
      fprintf (log_file," RPREF: 0x%02x   %03d\n",
               regs.rpref_reg, regs.rpref_reg);
      fprintf (log_file," GNREF: 0x%02x   %03d\n",
               regs.gnref_reg, regs.gnref_reg);
      fprintf (log_file," GPREF: 0x%02x   %03d\n",
               regs.gpref_reg, regs.gpref_reg);
      fprintf (log_file," BNREF: 0x%02x   %03d\n",
               regs.bnref_reg, regs.bnref_reg);
      fprintf (log_file," BPREF: 0x%02x   %03d\n",
               regs.bpref_reg, regs.bpref_reg);
      fprintf (log_file,"  CMD0: 0x%02x   %03d\n",
               regs.cmd0_reg, regs.cmd0_reg);
      fprintf (log_file,"  CMD1: 0x%02x   %03d\n",
               regs.cmd1_reg, regs.cmd1_reg);
    }
    if (log_action==2) break; else fprintf (log_file, "\n");

    /* write geometry of loaded FPGA-Timing to the log-file */
  case 3:
    if (ioctl (framegrabber, GIOC_C1_GET_FPGA, &geometry) < 0) {
      perror ("GIOC_C1_GET_FPGA");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Geometry)\n");
      error_count++;
    } else {
      fprintf (log_file, "%s : %d x %d @ %d\n", 
               geometry.mcs_name, 
               geometry.image_width, 
               geometry.image_height, 
               geometry.frames_per_sec);
    }
    if (log_action==3) break; else fprintf (log_file, "\n");

    /* write current aktive colormode of AM-CLR to the log-file */
  case 4:
    if (ioctl (framegrabber, GIOC_C1_GET_COLORMODE, &value) < 0) {
      perror ("GIOC_C1_GET_COLORMODE");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Colormode)\n");
      error_count++;
    } else {
      fprintf (log_file, "Colormode : %d\n", value);
    } 
    if (log_action==4) break; else fprintf (log_file, "\n");

    /* write Camera Configuration to the log-file */
  case 5:
    if (ioctl (framegrabber, GIOC_GE_GET_CAMCNF, &camcnf) < 0) {
      perror ("GIOC_GE_GET_CAMCNF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Camera Config)\n");
      error_count++;;
    } else {     
      fprintf (log_file, "CAMERA CONTROL : Super VHS Select : %d\n",
        camcnf.c1_cam.svhs);
      fprintf (log_file, "CAMERA CONTROL : Trigger 0 Enable : %d\n", 
        camcnf.c1_cam.trgen0);
      fprintf (log_file, "CAMERA CONTROL : Trigger 1 Enable : %d\n",
        camcnf.c1_cam.trgen1);
      fprintf (log_file, "CAMERA CONTROL : Trigger Polarity : %d\n", 
        camcnf.c1_cam.trgpol);
      fprintf (log_file, "MODULE CONTROL :       Sync Type Select : %d\n", 
        camcnf.c1_cam.synctyp);
      fprintf (log_file, "MODULE CONTROL :   Pixel Clock Polarity : %d\n", 
        camcnf.c1_cam.pclkpol);
      fprintf (log_file, "MODULE CONTROL : Low-Pass Filter Enable : %d\n",
        camcnf.c1_cam.lowpass);
      fprintf (log_file, "   ADC CONTROL :   Sync Strip Level : %d\n",
        camcnf.c1_cam.islvl);
      fprintf (log_file, "   ADC CONTROL :  Color Mode Select : %d\n",
        camcnf.c1_cam.color);
      fprintf (log_file, "   ADC CONTROL : Sync Source Select : %d\n", 
        camcnf.c1_cam.syncsel);
      fprintf (log_file, "     B LUT ADR :  Bypass Output LUT : %d\n",
        camcnf.c1_cam.bypass_olut);
    }
    if (log_action==5) break; else fprintf (log_file, "\n");

    /* write number of current aktive camera port to the log-file */
  case 6:
    if (ioctl (framegrabber, GIOC_GE_GET_CAMERA, &value) < 0) {
      perror ("GIOC_GE_GET_CAMERA");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Camera)\n");
      error_count++;
    } else {
      fprintf (log_file, "Camera : %d\n", value);
    }
    if (log_action==6) break; else fprintf (log_file, "\n");

    /* write neg. and pos. reference voltages to the log-file */
  case 7:
    ref.ref_color = 'R';
    ref.ref_sign = 'N';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Red Negative Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, "  Red Negative Reference : %d\n", ref.ref_value);
    }
    ref.ref_sign = 'P';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Red Positive Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, "  Red Positive Reference : %d\n", ref.ref_value);
    }
    ref.ref_color = 'G';
    ref.ref_sign = 'N';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Green Negative Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, "Green Negative Reference : %d\n", ref.ref_value);
    }
    ref.ref_sign = 'P';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Green Positive Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, "Green Positive Reference : %d\n", ref.ref_value);
    }
    ref.ref_color = 'B';
    ref.ref_sign = 'N';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Blue Negative Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, " Blue Negative Reference : %d\n", ref.ref_value);
    }
    ref.ref_sign = 'P';
    if (ioctl (framegrabber, GIOC_C1_GET_REF, &ref) < 0) {
      perror ("GIOC_C1_GET_REF");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Blue Positive Reference)\n");
      error_count++;
    } else {
      fprintf (log_file, " Blue Positive Reference : %d\n", ref.ref_value);
    }
    if (log_action==7) break; else fprintf (log_file, "\n");

    /* write Color Space Converter to the log-file */
  case 8:
    if (ioctl (framegrabber, GIOC_C1_GET_CSC, &csc) < 0) {
      perror ("GIOC_C1_GET_CSC");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Color Space Converter)\n");
      error_count++;
    } else {
      fprintf (log_file, "  CSC input format : %d\n", csc.csc_in);
      fprintf (log_file, " CSC output format : %d\n", csc.csc_out);
      fprintf (log_file, " CSC shift control : %d\n", csc.csc_shift);
      for (i=0; i<9; i++) {
        fprintf (log_file, "   CSC matrix[%d] : %f\n", i+1, csc.csc_matrix[i]);
      }
      fprintf (log_file, "CSC over/underflow : %d\n", csc.csc_overflow);
      fprintf (log_file,          "CSC error : %d\n", csc.csc_error);        
    }
    if (log_action==8) break; else fprintf (log_file, "\n");

    /* write Input LUT A, B and C to the log-file */
  case 9:
    ilut.ilut_name = 'A';        
    if (ioctl (framegrabber, GIOC_C1_GET_ILUT, &ilut) < 0) {
      perror ("GIOC_C1_GET_ILUT");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Input Lookup Table: %c)\n",
        ilut.ilut_name);
      error_count++;
    } else {
      fprintf (log_file, "Input Lookup Table: %c\n", ilut.ilut_name);
      for (i=0; i<16; i++) {
        for (k=0; k<16; k++) { 
          fprintf (log_file, "%03d ", ilut.ilut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    fprintf (log_file, "\n");
    ilut.ilut_name = 'B';
    if (ioctl (framegrabber, GIOC_C1_GET_ILUT, &ilut) < 0) {
      perror ("GIOC_C1_GET_ILUT");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Input Lookup Table: %c)\n",
        ilut.ilut_name);
      error_count++;
    } else {
      fprintf (log_file, "Input Lookup Table: %c\n", ilut.ilut_name);
      for (i=0; i<16; i++) {
        for (k=0; k<16; k++) { 
          fprintf (log_file, "%03d ", ilut.ilut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    fprintf (log_file, "\n");
    ilut.ilut_name = 'C';
    if (ioctl (framegrabber, GIOC_C1_GET_ILUT, &ilut) < 0) {
      perror ("GIOC_C1_GET_ILUT");
      fprintf (log_file, "!! IOCTL-ERROR !!  (Input Lookup Table: %c)\n",
        ilut.ilut_name);
      error_count++;
    } else {
      fprintf (log_file, "Input Lookup Table: %c\n", ilut.ilut_name);
      for (i=0; i<16; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", ilut.ilut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    if (log_action==9) break; else fprintf (log_file, "\n");

    /* write Output LUT A, B and C to the log-file */
  case 10:
    olut.olut_name = 'A';
    olut.olut_section = 0;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 1;
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 2;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 3;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    fprintf (log_file, "\n");
    olut.olut_name = 'B';
    olut.olut_section = 0; 
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c)\n", olut.olut_name);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c\n", olut.olut_name);
      for (i=0; i<128; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    fprintf (log_file, "\n");
    olut.olut_name = 'C';
    olut.olut_section = 0;
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 1;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 2;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }
    olut.olut_section = 3;  
    if (ioctl (framegrabber, GIOC_C1_GET_OLUT, &olut) < 0) {
      perror ("GIOC_C1_GET_OLUT");
      fprintf (log_file, 
        "!! IOCTL-ERROR !!  (Output Lookup Table: %c;  Section: %d)\n",
        olut.olut_name, olut.olut_section);
      error_count++;
    } else {
      fprintf (log_file, "Output Lookup Table: %c;  Section: %d\n", 
        olut.olut_name, olut.olut_section);
      for (i=0; i<512; i++) {
        for (k=0; k<16; k++) {
          fprintf (log_file, "%03d ", olut.olut_data[16*i+k]);
        }
        fprintf (log_file, "\n");
      }
    }   
    break;   
  }

  fclose (log_file);
  if (error_count != 0) return (-1);
  return (0);
}

/*
 * int fpga_load (struct c1_fpga_t *)
 *
 * read the FPGA-Timing from the proper ASCII-File with error checking,
 * calculate for every line the checksum and convert it into binary format
 */

int 
fpga_load (struct c1_fpga_t *fpga) {

  FILE *file;
  int number, high, low, record, data;
  int line_count, data_count, fpga_count=0;
  char *home, *pos, line[80];
  unsigned char checksum;

  /* get path of mcs file for open */

  if ((home = getenv ("HOME")) == NULL) {
    fprintf (stderr, "can't find environment variable 'HOME'.\n");
    return (-1);
  }

  strcpy (line, home);
  strcat (line, "/ITI/timing/");
  strcat (line, fpga->fpga_name);

  if ((file = fopen (line, "r")) == NULL) {
    fprintf (stderr, "can't open '%s'.\n", line);
    return (-1);
  }

  /* read mcs file and do error checking */

  for (line_count=0; fgets (line, 80, file) != NULL; line_count++) { 
    checksum = 0;
    pos = line;

    if (*pos != ':') {
      fprintf (stderr, "can't find ':' in file '%s' at line %d.\n", 
        fpga->fpga_name, line_count);
      return (-1);
    } 
    else pos++;

    /* read number of data bytes */

    if ((number = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);
      return (-1);
    }
    else {   
      pos += 2;     
      checksum += (unsigned char) number;
    }

    /* read high byte of addres */

    if ((high = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) high;
    }

    /* read low byte of addres */

    if ((low = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) low;
    }

    /* read record type */

    if ((record = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);
      return (-1);
    }
    else {
      pos += 2;
      checksum += (unsigned char) record;
    }
   
    /* read data bytes */
   
    for (data_count=0; data_count<number; data_count++) {
      if ((data = ascii2decimal (pos)) < 0) {    
        fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
          fpga->fpga_name, line_count);
        return (-1);
      }
      else {    
        if (record == 0) fpga->fpga_data[fpga_count++] = data;
        pos += 2;
        checksum += (unsigned char) data;
      }
    }

    /* read checksum byte */
 
    if ((data = ascii2decimal (pos)) < 0) {    
      fprintf (stderr, "illegal character found in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);
      return (-1);
    }
    else if ((unsigned char)(0x100 - checksum) != (unsigned char)data) {
      fprintf (stderr, "checksum error in file '%s' at line %d.\n",
        fpga->fpga_name, line_count);   
      return (-1);
    }
  }

  fclose (file);
  return (0);
}

/*
 * int ascii2decimal (char *)
 *
 * convert tow  characters, which are following another and form a
 * hexadecimal number into a decimal value
 */

int 
ascii2decimal (char *pos) {

  int value;

  /* convert ASCII character to the first digit of a decimal value */

  if ((*pos >= '0') && (*pos <= '9'))
    value = *pos++ - '0';
  else if ((*pos >= 'A') && (*pos <= 'F'))
    value = *pos++ - 'A' + 10;
  else if ((*pos >= 'a') && (*pos <= 'f'))
    value = *pos++ - 'a' + 10;
  else
    return (-1);

  /* convert ASCII character to the second digit of a decimal value */

  if ((*pos >= '0') && (*pos <= '9'))
    value = (value << 4) + *pos - '0';
  else if ((*pos >= 'A') && (*pos <= 'F'))
    value = (value << 4) + *pos - 'A' + 10;
  else if ((*pos >= 'a') && (*pos <= 'f'))
    value = (value << 4) + *pos - 'a' + 10;
  else
    return (-1);

  return value;
}

/*
 * void rgb2image (XImage *, u_long *, int)
 *
 * convert a NonInterlace TrueColor RGB image from framegrabber into
 * the X11 XImage format with notice of current X-Server depth
 */

void 
rgb2image (XImage *ximage, u_long *image_rgb, int width, int height) {

  int i, image_size; 

  u_char  *image_8bit;
  u_short *image_16bit;
  u_long *image_32bit;

  /* ITI : |0000|0000|BBBB|BBBB|GGGG|GGGG|RRRR|RRRR| */
  image_size = width * height; 

  switch (convert_mode) {

  /* 8 Bit : |00BB|GGRR| */
  case PSEUDO_8BIT:
    image_8bit = (u_char *) ximage->data;
    for (i=0; i<image_size; i++) {
      *(image_8bit) = ((*(image_rgb) & 0x0000C0) >> 6)      /* Red */
                    | ((*(image_rgb) & 0x00C000) >> 12)     /* Green */
                    | ((*(image_rgb) & 0xC00000) >> 18);    /* Blue */
      *(image_8bit++) = colortable[*(image_8bit)];
      image_rgb++;
    }
    break;

  /* 16 Bit : |RRRR|RGGG|GGGB|BBBB| */
  case RGB_16BIT:
    image_16bit = (u_short *) ximage->data;
    for (i=0; i<image_size; i++) {
      *(image_16bit++) = ((*(image_rgb) & 0x0000F8) << 8)   /* Red */ 
                       | ((*(image_rgb) & 0x00FC00) >> 5)   /* Green */
                       | ((*(image_rgb) & 0xF80000) >> 19); /* Blue */
      image_rgb++;
    }
    break;

  /* 16 Bit : |BBBB|BGGG|GGGR|RRRR| */
  case BGR_16BIT:
    image_16bit = (u_short *) ximage->data;
    for (i=0; i<image_size; i++) {
      *(image_16bit++) = ((*(image_rgb) & 0x0000F8) >> 3)   /* Red */
                       | ((*(image_rgb) & 0x00FC00) >> 5)   /* Green */
                       | ((*(image_rgb) & 0xF80000) >> 8);  /* Blue */
      image_rgb++; 
    }
    break;

  /* 32 Bit : |0000|0000|RRRR|RRRR|GGGG|GGGG|BBBB|BBBB| */
  case RGB_32BIT:
    image_32bit = (u_long *) ximage->data;
    for (i=0; i<image_size; i++) {
      *(image_32bit++) = ((*(image_rgb) & 0x0000FF) << 16)  /* Red */
                       |  (*(image_rgb) & 0x00FF00)         /* Green */
                       | ((*(image_rgb) & 0xFF0000) >> 16); /* Blue */
      image_rgb++; 
    }
    break;

  /* 32 Bit : |0000|0000|BBBB|BBBB|GGGG|GGGG|RRRR|RRRR| */
  case BGR_32BIT:
    image_32bit = (u_long *) ximage->data;
    for (i=0; i<image_size; i++) {
      *(image_32bit++) = *(image_rgb++);
    }
    break; 
  }
}

/*
 * void rgb_ilace2image (char *, u_long *, int, int)
 *
 * convert a Interlace TrueColor RGB image from framegrabber into
 * the X11 XImage format with notice of current X-Server depth
 * the even field must be stored first in framegrabber memory
 */

void 
rgb_ilace2image (XImage *image, u_long *image_rgb, int width, int height) {

  int i, line;

  u_char  *image_8bit;
  u_short *image_16bit;
  u_long *image_32bit;
  u_long *image_even, *image_odd;

  /* ITI : |0000|0000|BBBB|BBBB|GGGG|GGGG|RRRR|RRRR| */
  /*  first: Even Field */  
  image_even = image_rgb;
  /* second: Odd Field */   
  image_odd  = image_rgb + (width*height)/2;
   
  switch (convert_mode) {

  /* 8 Bit : |00BB|GGRR| */
  case PSEUDO_8BIT:
    image_8bit = (u_char *) image->data;
    for (i=0; i<height/2; i++) {
      for (line=0; line<width; line++) {
        *(image_8bit) = ((*(image_even) & 0x0000C0) >> 6)      /* Red */ 
                      | ((*(image_even) & 0x00C000) >> 12)     /* Green */
                      | ((*(image_even) & 0xC00000) >> 18);    /* Blue */
        *(image_8bit++) = colortable[*(image_8bit)];
        image_even++; 
      }
      for (line=0; line<width; line++) {
        *(image_8bit) = ((*(image_odd) & 0x0000C0) >> 6)       /* Red */ 
                      | ((*(image_odd) & 0x00C000) >> 12)      /* Green */
                      | ((*(image_odd) & 0xC00000) >> 18);     /* Blue */
        *(image_8bit++) = colortable[*(image_8bit)];
        image_odd++; 
      }
    }
    break;

  /* 16 Bit : |RRRRR|RGGG|GGGB|BBBB| */
  case RGB_16BIT:
    image_16bit = (u_short *) image->data;
    for (i=0; i<height/2; i++) {
      for (line=0; line<width; line++) {
        *(image_16bit++) = ((*(image_even) & 0x0000F8) << 8)   /* Red */ 
                         | ((*(image_even) & 0x00FC00) >> 5)   /* Green */
                         | ((*(image_even) & 0xF80000) >> 19); /* Blue */
        image_even++; 
      }
      for (line=0; line<width; line++) {
        *(image_16bit++) = ((*(image_odd) & 0x0000F8) << 8)    /* Red */ 
                         | ((*(image_odd) & 0x00FC00) >> 5)    /* Green */
                         | ((*(image_odd) & 0xF80000) >> 19);  /* Blue */
        image_odd++; 
      }
    }
    break;

  /* 16 Bit : |BBBB|BGGG|GGGR|RRRR| */
  case BGR_16BIT:
    image_16bit = (u_short *) image->data;            
    for (i=0; i<height/2; i++) {
      for (line=0; line<width; line++) {
        *(image_16bit++) = ((*(image_even) & 0x0000F8) >> 3)   /* Red */ 
                         | ((*(image_even) & 0x00FC00) >> 5)   /* Green */
                         | ((*(image_even) & 0xF80000) >> 8);  /* Blue */
        image_even++; 
      }
      for (line=0; line<width; line++) {
        *(image_16bit++) = ((*(image_odd) & 0x0000F8) >> 3)    /* Red */ 
                         | ((*(image_odd) & 0x00FC00) >> 5)    /* Green */
                         | ((*(image_odd) & 0xF80000) >> 8);   /* Blue */
        image_odd++; 
      }
    }
    break;

  /* 32 Bit : |0000|0000|RRRR|RRRR|GGGG|GGGG|BBBB|BBBB| */
  case RGB_32BIT:
    image_32bit = (u_long *) image->data;
    for (i=0; i<height/2; i++) {
      for (line=0; line<width; line++) {
        *(image_32bit++) = ((*(image_even) & 0x0000FF) << 16)  /* Red */ 
                         |  (*(image_even) & 0x00FF00)         /* Green */
                         | ((*(image_even) & 0xFF0000) >> 16); /* Blue */
        image_even++; 
      }
      for (line=0; line<width; line++) {
        *(image_32bit++) = ((*(image_odd) & 0x0000FF) << 16)   /* Red */ 
                         |  (*(image_odd) & 0x00FF00)          /* Green */
                         | ((*(image_odd) & 0xFF0000) >> 16);  /* Blue */
        image_odd++; 
      }
    }
    break;

  /* 32 Bit : |0000|0000|BBBB|BBBB|GGGG|GGGG|RRRR|RRRR| */
  case BGR_32BIT:
    image_32bit = (u_long *) image->data;
    for (i=0; i<height/2; i++) {
      for (line=0; line<width; line++) {
        *(image_32bit++) = *(image_even++);
      }
      for (line=0; line<width; line++) {
        *(image_32bit++) = *(image_odd++);
      }
    }
    break;
  }
}
